// VRAMCopy.c
//
// Fast copy routines for copying data to/from VRAM (without DMA).
//
// See the header file for details

#if defined(__AMIGA__) || defined(__amigaos4__)
#include <exec/exec.h>
#include <expansion/expansion.h>

#include <proto/exec.h>
#include <proto/expansion.h>
#endif

#include "Types.h"

#include "VRAMCopy.h"

#define BLOCKSIZE 32 // The length of a cache-line in the G4 processor

#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
bool hasAltivec = false;
bool altivecChecked = false;

#define CHECKALTIVEC \
	{											\
		if(!altivecChecked)						\
    	{										\
    		hasAltivec = has_altivec();			\
    		altivecChecked = true;				\
    	}										\
	}

static bool has_altivec() 
{
    uint32 machineType;
    IExpansion->GetMachineInfoTags(GMIT_Machine, (ULONG)&machineType, TAG_DONE);
    if(machineType == MACHINETYPE_AMIGAONE)
    {
        // AmigaOnes (the old ones) use the 60x bus protocol, which cannot handle
        // altivec accesses to cache-inhibited memory (such as VRAM). Hence,
        // we have to disable altivec
        return false;
    }
    
	uint32 vectorUnit;
	IExec->GetCPUInfoTags(GCIT_VectorUnit, &vectorUnit, TAG_DONE);

	if(vectorUnit == VECTORTYPE_ALTIVEC)
	{
	    return true;
	}
	else
	{
	    return false;
	}
}

#endif

void copyToVRAM(const void *src, void *dest, unsigned int numBytes)
{
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
    CHECKALTIVEC;
    
    if(hasAltivec)
    {
        copyToVRAMAltivec(src, dest, numBytes);
    }
    else
#endif
    {
        copyToVRAMNoAltivec(src, dest, numBytes);
    }
    
}

void copyFromVRAM(const void *src, void *dest, unsigned int numBytes)
{
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
    CHECKALTIVEC;
    
    if(hasAltivec)
    {
        copyFromVRAMAltivec(src, dest, numBytes);
    }
    else
#endif
    {
        copyFromVRAMNoAltivec(src, dest, numBytes);
    }
}

void copyToVRAMNoAltivec(const void *src, void *dest, unsigned int numBytes)
{
    // Advance until the destination address is 32-byte aligned
    if((uint32)dest & 1)
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        numBytes -= sizeof(uint8);
    }
    if((uint32)dest & 2 && numBytes >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        numBytes -= sizeof(uint16);
    }
    if((uint32)dest & 4 && numBytes >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        numBytes -= sizeof(uint32);
    }
    double *src64 = (double*)src;
    double *dest64 = (double*)dest;
    while((uint32)dest64 & 0x1F && numBytes >= sizeof(double))
    {
        *dest64 = *src64;
        ++src64;
        ++dest64;
        numBytes -= sizeof(double);
    }

    // Perform the main copy
    if(numBytes >= BLOCKSIZE)
    {
        // WARNING: BLOCKSIZE *MUST* be a power of two
        double *destEnd = dest64 + (numBytes & ~(BLOCKSIZE - 1)) / sizeof(double);
        double temp1, temp2;
        while(dest64 < destEnd)
        {
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
            __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64));
            __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64 + BLOCKSIZE));
#endif            
            temp1 = src64[0];
            temp2 = src64[1];
            src64 += 2;
            dest64[0] = temp1;
            dest64[1] = temp2;
            dest64 +=2;
            temp1 = src64[0];
            temp2 = src64[1];
            src64 += 2;
            dest64[0] = temp1;
            dest64[1] = temp2;
            dest64 +=2;
        }
    }

    // Copy the last few bytes
    unsigned int residue = numBytes % BLOCKSIZE;
    src = src64;
    dest = dest64;
    while(residue >= sizeof(double))
    {
        *((double*)dest) = *((double*)src);
        src = (void*)(((double*)src) + 1);
        dest = (void*)(((double*)dest) + 1);
        residue -= sizeof(double);
    }
    while(residue >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        residue -= sizeof(uint32);
    }
    while(residue >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        residue -= sizeof(uint16);
    }
    while(residue >= sizeof(uint8))
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        residue -= sizeof(uint8);
    }
}

void copyFromVRAMNoAltivec(const void *src, void *dest, unsigned int numBytes)
{
    bool destCacheAligned;

    // The most important thing to do is to make sure that the address on the graphics
    // card is at least 32-bit word aligned

    // Advance until the source address is 32-bit word aligned
    if((uint32)src & 1)
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        numBytes -= sizeof(uint8);
    }
    if((uint32)src & 2 && numBytes >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        numBytes -= sizeof(uint16);
    }


    double  *src64, *dest64;

    if(((uint32)dest & 3) == 0)
    {
        // Align the destination buffer to the cache boundary
        if((uint32)dest & 4 && numBytes >= sizeof(uint32))
        {
            *((uint32*)dest) = *((uint32*)src);
            src = (void*)(((uint32*)src) + 1);
            dest = (void*)(((uint32*)dest) + 1);
            numBytes -= sizeof(uint32);
        }

        src64 = (double*)src;
        dest64 = (double*)dest;

        while((uint32)dest64 & 0x1F && numBytes >= sizeof(double))
        {
            *dest64 = *src64;
            ++src64;
            ++dest64;
            numBytes -= sizeof(double);
        }

        destCacheAligned = true;
    }
    else
    { 
        // Can't make the destination buffer cache-aligned, so make the source
        // 64-bit word aligned
        if((uint32)src & 4 && numBytes >= sizeof(uint32))
        {
            *((uint32*)dest) = *((uint32*)src);
            src = (void*)(((uint32*)src) + 1);
            dest = (void*)(((uint32*)dest) + 1);
            numBytes -= sizeof(uint32);
        }

        src64 = (double*)src;
        dest64 = (double*)dest;
        destCacheAligned = false;
    }

    // Perform the main copy
    if(numBytes >= BLOCKSIZE)
    {
        if(destCacheAligned)
        {
            // WARNING: BLOCKSIZE *MUST* be a power of two
            double *srcEnd = src64 + (numBytes & ~(BLOCKSIZE - 1)) / sizeof(double);
            while(src64 < srcEnd)
            {
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
                __asm__ __volatile__ ("dcbz 0,%0" : : "r" (dest64));
#endif
                dest64[0] = src64[0];
                dest64[1] = src64[1];
                dest64[2] = src64[2];
                dest64[3] = src64[3];
                src64 += 4;
                dest64 += 4;
            }
        }
        else
        {
            // WARNING: BLOCKSIZE *MUST* be a power of two
            double *srcEnd = src64 + (numBytes & ~(BLOCKSIZE - 1)) / sizeof(double);
            while(src64 < srcEnd)
            {
                dest64[0] = src64[0];
                dest64[1] = src64[1];
                dest64[2] = src64[2];
                dest64[3] = src64[3];
                src64 += 4;
                dest64 += 4;
            }
        }
    }

    // Copy the last few bytes
    unsigned int residue = numBytes % BLOCKSIZE;
    src = src64;
    dest = dest64;
    while(residue >= sizeof(double))
    {
        *((double*)dest) = *((double*)src);
        src = (void*)(((double*)src) + 1);
        dest = (void*)(((double*)dest) + 1);
        residue -= sizeof(double);
    }
    while(residue >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        residue -= sizeof(uint32);
    }
    while(residue >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        residue -= sizeof(uint16);
    }
    while(residue >= sizeof(uint8))
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        residue -= sizeof(uint8);
    }
}

