// VRAMCopyAltivec.c
//
// Fast copy routines for copying data to/from VRAM (without DMA).
//
// See the header file for details

#include <altivec.h>

#include "Types.h"

#include "VRAMCopy.h"

#define BLOCKSIZE 32 // The length of a cache-line in the G4 processor

/** Loads data into an altivec float vector from a 16-byte aligned buffer.
 *
 * @param vec the vector that the data is being loaded into
 * @param data the source buffer
 */
#define LOAD_ALIGNED_VECTORF(vec, data)	\
	{												\
		vec = vec_ld(0, (float*)data);				\
	}
	
/** Loads data into an altivec float vector from a buffer that is NOT 16-byte
 * aligned.
 *
 * @param vec the vector that the data is being loaded into
 * @param data the source buffer
 */
#define LOAD_UNALIGNED_VECTORF(vec, data)	\
	{													\
		vector unsigned char MSQ, LSQ;					\
		vector unsigned char mask; 						\
		MSQ = vec_ld(0, (unsigned char*)data);			\
		LSQ = vec_ld(15, (unsigned char*)data);			\
		mask = vec_lvsl(0, (unsigned char*)data);		\
		vec = (vector float)vec_perm(MSQ, LSQ, mask);	\
	}

/** Stores data into an altivec float vector to a 16-byte aligned buffer.
 *
 * @param vec the vector that the data is being loaded into
 * @param data the output buffer
 */
#define STORE_ALIGNED_VECTORF(vec, data)	\
	{												\
		vec_st(vec, 0, (float*)data);				\
	}
	
/** Loads data into an altivec float vector to a buffer that is NOT 16-byte
 * aligned.
 *
 * NOTE: We are assuming (correctly) that the destination is 32-bit aligned
 *
 * @param vec the vector that the data is being loaded into
 * @param data the output buffer
 */
#define STORE_UNALIGNED_VECTORF(vec, data)	\
	{												\
		vector unsigned char MSQ, LSQ, edges;								\
		vector unsigned char edgeAlign, align;								\
		MSQ = vec_ld(0, (unsigned char*)data); 								\
 		LSQ = vec_ld(15, (unsigned char*)data);								\
		edgeAlign = vec_lvsl(0, (unsigned char*)data);						\
		edges = vec_perm(LSQ,MSQ,edgeAlign);								\
		align = vec_lvsr( 0, (unsigned char*)data);							\
		MSQ = vec_perm(edges,(vector unsigned char)vec,align);				\
		LSQ = vec_perm((vector unsigned char)vec,edges,align);				\
		vec_st( LSQ, 15, (unsigned char*)data);								\
		vec_st( MSQ, 0, (unsigned char*)data);								\
	}

void copyToVRAMAltivec(const void *src, void *dest, unsigned int numBytes)
{
    // Advance until the destination address is 32-byte aligned
    if((uint32)dest & 1)
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        numBytes -= sizeof(uint8);
    }
    if((uint32)dest & 2 && numBytes >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        numBytes -= sizeof(uint16);
    }
    if((uint32)dest & 4 && numBytes >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        numBytes -= sizeof(uint32);
    }
    double *src64 = (double*)src;
    double *dest64 = (double*)dest;
    while((uint32)dest64 & 0x1F && numBytes >= sizeof(double))
    {
        *dest64 = *src64;
        ++src64;
        ++dest64;
        numBytes -= sizeof(double);
    }

    // Perform the main copy
    if(numBytes >= BLOCKSIZE)
    {
        // WARNING: BLOCKSIZE *MUST* be a power of two
        double *destEnd = dest64 + (numBytes & ~(BLOCKSIZE - 1)) / sizeof(double);
        if(((uint32)src64 & 0xF) != 0)
        {
            // Source isn't correctly aligned for altivec
            // NOTE: The destination is correctly aligned due to the code above
            while(dest64 < destEnd)
            {
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
                __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64));
                __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64 + BLOCKSIZE));
#endif
				vector float vec;
                LOAD_UNALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
                LOAD_UNALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
            }
        }
        else
        {
            // Both source and destination are correctly aligned for altivec
            while(dest64 < destEnd)
            {
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
                __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64));
                __asm__ __volatile__ ("dcbt 0,%0" : : "r" (src64 + BLOCKSIZE));
#endif
				vector float vec;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
            }
        }
    }

    // Copy the last few bytes
    unsigned int residue = numBytes % BLOCKSIZE;
    src = src64;
    dest = dest64;
    while(residue >= sizeof(double))
    {
        *((double*)dest) = *((double*)src);
        src = (void*)(((double*)src) + 1);
        dest = (void*)(((double*)dest) + 1);
        residue -= sizeof(double);
    }
    while(residue >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        residue -= sizeof(uint32);
    }
    while(residue >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        residue -= sizeof(uint16);
    }
    while(residue >= sizeof(uint8))
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        residue -= sizeof(uint8);
    }
}

void copyFromVRAMAltivec(const void *src, void *dest, unsigned int numBytes)
{
    BOOL destCacheAligned;

    // Advance until the source address is aligned for altivec operations
    if((uint32)src & 1)
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        numBytes -= sizeof(uint8);
    }
    if((uint32)src & 2 && numBytes >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        numBytes -= sizeof(uint16);
    }
    if((uint32)src & 4 && numBytes >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        numBytes -= sizeof(uint32);
    }
    double *src64 = (double*)src;
    double *dest64 = (double*)dest;
    while((uint32)src64 & 0xF && numBytes >= sizeof(double))
    {
        *dest64 = *src64;
        ++src64;
        ++dest64;
        numBytes -= sizeof(double);
    }

    if(((uint32)dest64 & 0xF) == 0)
    {
        // Make the destination address cache-aligned
        while((uint32)dest64 & 0x1F && numBytes >= sizeof(double))
        {
            *dest64 = *src64;
            ++src64;
            ++dest64;
            numBytes -= sizeof(double);
        }
        destCacheAligned = TRUE;
    }
    else
    {
        destCacheAligned = FALSE;
    }

    // Perform the main copy
    if(numBytes >= BLOCKSIZE)
    {
        // WARNING: BLOCKSIZE *MUST* be a power of two
        double *srcEnd = src64 + (numBytes & ~(BLOCKSIZE - 1)) / sizeof(double);
        
        if(((uint32)dest64 & 0xF) != 0)
        {
            // Destination isn't correctly aligned for altivec
            // NOTE: The source is correctly aligned due to the code above
            while(src64 < srcEnd)
            {
				vector float vec;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_UNALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_UNALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
            }
        }
        else
        {
            // Both source and destination are correctly aligned for altivec
            while(src64 < srcEnd)
            {
#if defined(__POWERPC__) || defined (_M_PPC) || defined(__powerpc__)
                if(destCacheAligned)
               {
                    __asm__ __volatile__ ("dcbz 0,%0" : : "r" (dest64));
                }
#endif
				vector float vec;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
                LOAD_ALIGNED_VECTORF(vec, src64);
                STORE_ALIGNED_VECTORF(vec, dest64);
                src64 += 2;
                dest64 += 2;
            }
        }
    }

    // Copy the last few bytes
    unsigned int residue = numBytes % BLOCKSIZE;
    src = src64;
    dest = dest64;
    while(residue >= sizeof(double))
    {
        *((double*)dest) = *((double*)src);
        src = (void*)(((double*)src) + 1);
        dest = (void*)(((double*)dest) + 1);
        residue -= sizeof(double);
    }
    while(residue >= sizeof(uint32))
    {
        *((uint32*)dest) = *((uint32*)src);
        src = (void*)(((uint32*)src) + 1);
        dest = (void*)(((uint32*)dest) + 1);
        residue -= sizeof(uint32);
    }
    while(residue >= sizeof(uint16))
    {
        *((uint16*)dest) = *((uint16*)src);
        src = (void*)(((uint16*)src) + 1);
        dest = (void*)(((uint16*)dest) + 1);
        residue -= sizeof(uint16);
    }
    while(residue >= sizeof(uint8))
    {
        *((uint8*)dest) = *((uint8*)src);
        src = (void*)(((uint8*)src) + 1);
        dest = (void*)(((uint8*)dest) + 1);
        residue -= sizeof(uint8);
    }
}

